@@ -11,7 +11,7 @@
 #include <math.h>
 
 #include <clutter/x11/clutter-x11.h>
-#include <cogl/cogl-texture-pixmap-x11.h>
+#include <cogl/winsys/cogl-texture-pixmap-x11.h>
 #include <gdk/gdk.h> /* for gdk_rectangle_union() */
 #include <string.h>
 
@@ -20,14 +20,16 @@
 #include "frame.h"
 #include <meta/window.h>
 #include <meta/meta-shaped-texture.h>
+#include <meta/meta-enum-types.h>
+#include <meta/meta-shadow-factory.h>
 
+#include "clutter/clutter-mutter.h"
 #include "compositor-private.h"
 #include "meta-shaped-texture-private.h"
-#include "meta-shadow-factory-private.h"
 #include "meta-window-actor-private.h"
 #include "meta-texture-rectangle.h"
 #include "region-utils.h"
-#include "meta-monitor-manager.h"
+#include "meta-monitor-manager-private.h"
 #include "meta-cullable.h"
 
 #include "meta-surface-actor.h"
@@ -74,6 +76,8 @@ struct _MetaWindowActorPrivate
   MetaWindowShape  *shadow_shape;
   char *            shadow_class;
 
+  MetaShadowMode    shadow_mode;
+
   guint             send_frame_messages_timer;
   gint64            frame_drawn_time;
 
@@ -87,8 +91,7 @@ struct _MetaWindowActorPrivate
    */
   gint              minimize_in_progress;
   gint              unminimize_in_progress;
-  gint              maximize_in_progress;
-  gint              unmaximize_in_progress;
+  gint              size_change_in_progress;
   gint              map_in_progress;
   gint              destroy_in_progress;
 
@@ -110,8 +113,6 @@ struct _MetaWindowActorPrivate
 
   guint		    needs_destroy	   : 1;
 
-  guint             no_shadow              : 1;
-
   guint             updates_frozen         : 1;
   guint             first_frame_state      : 2; /* FirstFrameState */
 };
@@ -147,7 +148,7 @@ static guint signals[LAST_SIGNAL] = { 0 };
 enum
 {
   PROP_META_WINDOW = 1,
-  PROP_NO_SHADOW,
+  PROP_SHADOW_MODE,
   PROP_SHADOW_CLASS
 };
 
@@ -245,14 +246,15 @@ meta_window_actor_class_init (MetaWindowActorClass *klass)
                                    PROP_META_WINDOW,
                                    pspec);
 
-  pspec = g_param_spec_boolean ("no-shadow",
-                                "No shadow",
-                                "Do not add shaddow to this window",
-                                FALSE,
-                                G_PARAM_READWRITE);
+  pspec = g_param_spec_enum ("shadow-mode",
+                             "Shadow mode",
+                             "Decides when to paint shadows",
+                             META_TYPE_SHADOW_MODE,
+                             META_SHADOW_MODE_AUTO,
+                             G_PARAM_READWRITE);
 
   g_object_class_install_property (object_class,
-                                   PROP_NO_SHADOW,
+                                   PROP_SHADOW_MODE,
                                    pspec);
 
   pspec = g_param_spec_string ("shadow-class",
@@ -347,6 +349,21 @@ meta_window_actor_freeze (MetaWindowActor *self)
   priv->freeze_count ++;
 }
 
+static void
+meta_window_actor_sync_thawed_state (MetaWindowActor *self)
+{
+  MetaWindowActorPrivate *priv = self->priv;
+
+  if (priv->first_frame_state == INITIALLY_FROZEN)
+    priv->first_frame_state = DRAWING_FIRST_FRAME;
+
+  if (priv->surface)
+    meta_surface_actor_set_frozen (priv->surface, FALSE);
+
+  /* We sometimes ignore moves and resizes on frozen windows */
+  meta_window_actor_sync_actor_geometry (self, FALSE);
+}
+
 static void
 meta_window_actor_thaw (MetaWindowActor *self)
 {
@@ -359,14 +376,11 @@ meta_window_actor_thaw (MetaWindowActor *self)
   if (priv->freeze_count > 0)
     return;
 
-  if (priv->first_frame_state == INITIALLY_FROZEN)
-    priv->first_frame_state = DRAWING_FIRST_FRAME;
+  /* We still might be frozen due to lack of a MetaSurfaceActor */
+  if (is_frozen (self))
+    return;
 
-  if (priv->surface)
-    meta_surface_actor_set_frozen (priv->surface, FALSE);
-
-  /* We sometimes ignore moves and resizes on frozen windows */
-  meta_window_actor_sync_actor_geometry (self, FALSE);
+  meta_window_actor_sync_thawed_state (self);
 
   /* We do this now since we might be going right back into the
    * frozen state */
@@ -399,14 +413,12 @@ set_surface (MetaWindowActor  *self,
                                                 G_CALLBACK (surface_size_changed), self);
       clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (priv->surface));
 
-      /* If the previous surface actor was frozen, start out
-       * frozen as well... */
-      meta_surface_actor_set_frozen (priv->surface, priv->freeze_count > 0);
-
-      if (!is_frozen (self) && priv->first_frame_state == INITIALLY_FROZEN)
-        priv->first_frame_state = DRAWING_FIRST_FRAME;
-
       meta_window_actor_update_shape (self);
+
+      if (is_frozen (self))
+        meta_surface_actor_set_frozen (priv->surface, TRUE);
+      else
+        meta_window_actor_sync_thawed_state (self);
     }
 }
 
@@ -510,14 +522,14 @@ meta_window_actor_set_property (GObject      *object,
       g_signal_connect_object (priv->window, "notify::appears-focused",
                                G_CALLBACK (window_appears_focused_notify), self, 0);
       break;
-    case PROP_NO_SHADOW:
+    case PROP_SHADOW_MODE:
       {
-        gboolean newv = g_value_get_boolean (value);
+        MetaShadowMode newv = g_value_get_enum (value);
 
-        if (newv == priv->no_shadow)
+        if (newv == priv->shadow_mode)
           return;
 
-        priv->no_shadow = newv;
+        priv->shadow_mode = newv;
 
         meta_window_actor_invalidate_shadow (self);
       }
@@ -554,8 +566,8 @@ meta_window_actor_get_property (GObject      *object,
     case PROP_META_WINDOW:
       g_value_set_object (value, priv->window);
       break;
-    case PROP_NO_SHADOW:
-      g_value_set_boolean (value, priv->no_shadow);
+    case PROP_SHADOW_MODE:
+      g_value_set_enum (value, priv->shadow_mode);
       break;
     case PROP_SHADOW_CLASS:
       g_value_set_string (value, priv->shadow_class);
@@ -611,18 +623,6 @@ meta_window_actor_get_shape_bounds (MetaWindowActor       *self,
   MetaWindowActorPrivate *priv = self->priv;
 
   cairo_region_get_extents (priv->shape_region, bounds);
-
-#ifdef HAVE_WAYLAND
-  if (META_IS_SURFACE_ACTOR_WAYLAND (priv->surface))
-    {
-      double scale = priv->surface ?
-                     meta_surface_actor_wayland_get_scale (META_SURFACE_ACTOR_WAYLAND (priv->surface)) : 1.;
-      bounds->x *= scale;
-      bounds->y *= scale;
-      bounds->width *= scale;
-      bounds->height *= scale;
-    }
-#endif
 }
 
 static void
@@ -670,6 +670,8 @@ static void
 assign_frame_counter_to_frames (MetaWindowActor *self)
 {
   MetaWindowActorPrivate *priv = self->priv;
+  MetaCompositor *compositor = priv->compositor;
+  ClutterStage *stage = CLUTTER_STAGE (compositor->stage);
   GList *l;
 
   /* If the window is obscured, then we're expecting to deal with sending
@@ -683,10 +685,7 @@ assign_frame_counter_to_frames (MetaWindowActor *self)
       FrameData *frame = l->data;
 
       if (frame->frame_counter == -1)
-        {
-          CoglOnscreen *onscreen = COGL_ONSCREEN (cogl_get_draw_framebuffer());
-          frame->frame_counter = cogl_onscreen_get_frame_counter (onscreen);
-        }
+        frame->frame_counter = clutter_stage_get_frame_counter (stage);
     }
 }
 
@@ -802,8 +801,10 @@ meta_window_actor_has_shadow (MetaWindowActor *self)
 {
   MetaWindowActorPrivate *priv = self->priv;
 
-  if (priv->no_shadow)
+  if (priv->shadow_mode == META_SHADOW_MODE_FORCED_OFF)
     return FALSE;
+  if (priv->shadow_mode == META_SHADOW_MODE_FORCED_ON)
+    return TRUE;
 
   /* Leaving out shadows for maximized and fullscreen windows is an effeciency
    * win and also prevents the unsightly effect of the shadow of maximized
@@ -827,21 +828,23 @@ meta_window_actor_has_shadow (MetaWindowActor *self)
     return TRUE;
 
   /*
-   * Do not add shadows to non-opaque windows; eventually we should generate
-   * a shadow from the input shape for such windows.
+   * Do not add shadows to non-opaque (ARGB32) windows, as we can't easily
+   * generate shadows for them.
    */
   if (is_non_opaque (self))
     return FALSE;
 
   /*
-   * Add shadows to override redirect windows on X11 unless the toolkit
-   * indicates that it is handling shadows itself (e.g., Gtk menus).
+   * If a window specifies that it has custom frame extents, that likely
+   * means that it is drawing a shadow itself. Don't draw our own.
    */
-  if (priv->window->override_redirect &&
-      !priv->window->has_custom_frame_extents)
-    return TRUE;
+  if (priv->window->has_custom_frame_extents)
+    return FALSE;
 
-  return FALSE;
+  /*
+   * Generate shadows for all other windows.
+   */
+  return TRUE;
 }
 
 /**
@@ -1036,20 +1039,18 @@ gboolean
 meta_window_actor_effect_in_progress (MetaWindowActor *self)
 {
   return (self->priv->minimize_in_progress ||
-	  self->priv->maximize_in_progress ||
-	  self->priv->unmaximize_in_progress ||
+	  self->priv->size_change_in_progress ||
 	  self->priv->map_in_progress ||
 	  self->priv->destroy_in_progress);
 }
 
 static gboolean
-is_freeze_thaw_effect (gulong event)
+is_freeze_thaw_effect (MetaPluginEffect event)
 {
   switch (event)
   {
   case META_PLUGIN_DESTROY:
-  case META_PLUGIN_MAXIMIZE:
-  case META_PLUGIN_UNMAXIMIZE:
+  case META_PLUGIN_SIZE_CHANGE:
     return TRUE;
     break;
   default:
@@ -1058,8 +1059,8 @@ is_freeze_thaw_effect (gulong event)
 }
 
 static gboolean
-start_simple_effect (MetaWindowActor *self,
-                     gulong        event)
+start_simple_effect (MetaWindowActor  *self,
+                     MetaPluginEffect  event)
 {
   MetaWindowActorPrivate *priv = self->priv;
   MetaCompositor *compositor = priv->compositor;
@@ -1068,6 +1069,8 @@ start_simple_effect (MetaWindowActor *self,
 
   switch (event)
   {
+  case META_PLUGIN_NONE:
+    return FALSE;
   case META_PLUGIN_MINIMIZE:
     counter = &priv->minimize_in_progress;
     break;
@@ -1080,8 +1083,7 @@ start_simple_effect (MetaWindowActor *self,
   case META_PLUGIN_DESTROY:
     counter = &priv->destroy_in_progress;
     break;
-  case META_PLUGIN_UNMAXIMIZE:
-  case META_PLUGIN_MAXIMIZE:
+  case META_PLUGIN_SIZE_CHANGE:
   case META_PLUGIN_SWITCH_WORKSPACE:
     g_assert_not_reached ();
     break;
@@ -1123,10 +1125,11 @@ meta_window_actor_after_effects (MetaWindowActor *self)
 }
 
 void
-meta_window_actor_effect_completed (MetaWindowActor *self,
-                                    gulong           event)
+meta_window_actor_effect_completed (MetaWindowActor  *self,
+                                    MetaPluginEffect  event)
 {
   MetaWindowActorPrivate *priv   = self->priv;
+  gboolean inconsistent = FALSE;
 
   /* NB: Keep in mind that when effects get completed it possible
    * that the corresponding MetaWindow may have be been destroyed.
@@ -1134,6 +1137,8 @@ meta_window_actor_effect_completed (MetaWindowActor *self,
 
   switch (event)
   {
+  case META_PLUGIN_NONE:
+    break;
   case META_PLUGIN_MINIMIZE:
     {
       priv->minimize_in_progress--;
@@ -1141,6 +1146,7 @@ meta_window_actor_effect_completed (MetaWindowActor *self,
 	{
 	  g_warning ("Error in minimize accounting.");
 	  priv->minimize_in_progress = 0;
+          inconsistent = TRUE;
 	}
     }
     break;
@@ -1151,6 +1157,7 @@ meta_window_actor_effect_completed (MetaWindowActor *self,
        {
          g_warning ("Error in unminimize accounting.");
          priv->unminimize_in_progress = 0;
+         inconsistent = TRUE;
        }
     }
     break;
@@ -1165,6 +1172,7 @@ meta_window_actor_effect_completed (MetaWindowActor *self,
       {
 	g_warning ("Error in map accounting.");
 	priv->map_in_progress = 0;
+        inconsistent = TRUE;
       }
     break;
   case META_PLUGIN_DESTROY:
@@ -1174,22 +1182,16 @@ meta_window_actor_effect_completed (MetaWindowActor *self,
       {
 	g_warning ("Error in destroy accounting.");
 	priv->destroy_in_progress = 0;
+        inconsistent = TRUE;
       }
     break;
-  case META_PLUGIN_UNMAXIMIZE:
-    priv->unmaximize_in_progress--;
-    if (priv->unmaximize_in_progress < 0)
+  case META_PLUGIN_SIZE_CHANGE:
+    priv->size_change_in_progress--;
+    if (priv->size_change_in_progress < 0)
       {
-	g_warning ("Error in unmaximize accounting.");
-	priv->unmaximize_in_progress = 0;
-      }
-    break;
-  case META_PLUGIN_MAXIMIZE:
-    priv->maximize_in_progress--;
-    if (priv->maximize_in_progress < 0)
-      {
-	g_warning ("Error in maximize accounting.");
-	priv->maximize_in_progress = 0;
+	g_warning ("Error in size change accounting.");
+	priv->size_change_in_progress = 0;
+        inconsistent = TRUE;
       }
     break;
   case META_PLUGIN_SWITCH_WORKSPACE:
@@ -1197,7 +1199,7 @@ meta_window_actor_effect_completed (MetaWindowActor *self,
     break;
   }
 
-  if (is_freeze_thaw_effect (event))
+  if (is_freeze_thaw_effect (event) && !inconsistent)
     meta_window_actor_thaw (self);
 
   if (!meta_window_actor_effect_in_progress (self))
@@ -1300,7 +1302,7 @@ meta_window_actor_show (MetaWindowActor   *self,
 {
   MetaWindowActorPrivate *priv = self->priv;
   MetaCompositor *compositor = priv->compositor;
-  gulong event = 0;
+  MetaPluginEffect event;
 
   g_return_if_fail (!priv->visible);
 
@@ -1315,14 +1317,13 @@ meta_window_actor_show (MetaWindowActor   *self,
       event = META_PLUGIN_UNMINIMIZE;
       break;
     case META_COMP_EFFECT_NONE:
+      event = META_PLUGIN_NONE;
       break;
-    case META_COMP_EFFECT_DESTROY:
-    case META_COMP_EFFECT_MINIMIZE:
+    default:
       g_assert_not_reached();
     }
 
   if (compositor->switch_workspace_in_progress ||
-      event == 0 ||
       !start_simple_effect (self, event))
     {
       clutter_actor_show (CLUTTER_ACTOR (self));
@@ -1335,7 +1336,7 @@ meta_window_actor_hide (MetaWindowActor *self,
 {
   MetaWindowActorPrivate *priv = self->priv;
   MetaCompositor *compositor = priv->compositor;
-  gulong event = 0;
+  MetaPluginEffect event;
 
   g_return_if_fail (priv->visible);
 
@@ -1357,70 +1358,32 @@ meta_window_actor_hide (MetaWindowActor *self,
       event = META_PLUGIN_MINIMIZE;
       break;
     case META_COMP_EFFECT_NONE:
+      event = META_PLUGIN_NONE;
       break;
-    case META_COMP_EFFECT_UNMINIMIZE:
-    case META_COMP_EFFECT_CREATE:
+    default:
       g_assert_not_reached();
     }
 
-  if (event == 0 ||
-      !start_simple_effect (self, event))
+  if (!start_simple_effect (self, event))
     clutter_actor_hide (CLUTTER_ACTOR (self));
 }
 
 void
-meta_window_actor_maximize (MetaWindowActor    *self,
-                            MetaRectangle      *old_rect,
-                            MetaRectangle      *new_rect)
+meta_window_actor_size_change (MetaWindowActor    *self,
+                               MetaSizeChange      which_change,
+                               MetaRectangle      *old_frame_rect,
+                               MetaRectangle      *old_buffer_rect)
 {
   MetaWindowActorPrivate *priv = self->priv;
   MetaCompositor *compositor = priv->compositor;
 
-  /* The window has already been resized (in order to compute new_rect),
-   * which by side effect caused the actor to be resized. Restore it to the
-   * old size and position */
-  clutter_actor_set_position (CLUTTER_ACTOR (self), old_rect->x, old_rect->y);
-  clutter_actor_set_size (CLUTTER_ACTOR (self), old_rect->width, old_rect->height);
-
-  self->priv->maximize_in_progress++;
+  self->priv->size_change_in_progress++;
   meta_window_actor_freeze (self);
 
-  if (!meta_plugin_manager_event_maximize (compositor->plugin_mgr,
-                                           self,
-                                           META_PLUGIN_MAXIMIZE,
-                                           new_rect->x, new_rect->y,
-                                           new_rect->width, new_rect->height))
-
+  if (!meta_plugin_manager_event_size_change (compositor->plugin_mgr, self,
+                                              which_change, old_frame_rect, old_buffer_rect))
     {
-      self->priv->maximize_in_progress--;
-      meta_window_actor_thaw (self);
-    }
-}
-
-void
-meta_window_actor_unmaximize (MetaWindowActor   *self,
-                              MetaRectangle     *old_rect,
-                              MetaRectangle     *new_rect)
-{
-  MetaWindowActorPrivate *priv = self->priv;
-  MetaCompositor *compositor = priv->compositor;
-
-  /* The window has already been resized (in order to compute new_rect),
-   * which by side effect caused the actor to be resized. Restore it to the
-   * old size and position */
-  clutter_actor_set_position (CLUTTER_ACTOR (self), old_rect->x, old_rect->y);
-  clutter_actor_set_size (CLUTTER_ACTOR (self), old_rect->width, old_rect->height);
-
-  self->priv->unmaximize_in_progress++;
-  meta_window_actor_freeze (self);
-
-  if (!meta_plugin_manager_event_maximize (compositor->plugin_mgr,
-                                           self,
-                                           META_PLUGIN_UNMAXIMIZE,
-                                           new_rect->x, new_rect->y,
-                                           new_rect->width, new_rect->height))
-    {
-      self->priv->unmaximize_in_progress--;
+      self->priv->size_change_in_progress--;
       meta_window_actor_thaw (self);
     }
 }
@@ -1791,9 +1754,17 @@ build_and_scan_frame_mask (MetaWindowActor       *self,
     }
   else
     {
+      CoglError *error = NULL;
+
       mask_texture = COGL_TEXTURE (cogl_texture_2d_new_from_data (ctx, tex_width, tex_height,
                                                                   COGL_PIXEL_FORMAT_A_8,
-                                                                  stride, mask_data, NULL));
+                                                                  stride, mask_data, &error));
+
+      if (error)
+        {
+          g_warning ("Failed to allocate mask texture: %s", error->message);
+          cogl_error_free (error);
+        }
     }
 
   meta_shaped_texture_set_mask_texture (stex, mask_texture);
@@ -2074,13 +2045,13 @@ do_send_frame_timings (MetaWindowActor  *self,
 static void
 send_frame_timings (MetaWindowActor  *self,
                     FrameData        *frame,
-                    CoglFrameInfo    *frame_info,
+                    ClutterFrameInfo *frame_info,
                     gint64            presentation_time)
 {
   float refresh_rate;
   int refresh_interval;
 
-  refresh_rate = cogl_frame_info_get_refresh_rate (frame_info);
+  refresh_rate = frame_info->refresh_rate;
   /* 0.0 is a flag for not known, but sanity-check against other odd numbers */
   if (refresh_rate >= 1.0)
     refresh_interval = (int) (0.5 + 1000000 / refresh_rate);
@@ -2091,9 +2062,9 @@ send_frame_timings (MetaWindowActor  *self,
 }
 
 void
-meta_window_actor_frame_complete (MetaWindowActor *self,
-                                  CoglFrameInfo   *frame_info,
-                                  gint64           presentation_time)
+meta_window_actor_frame_complete (MetaWindowActor  *self,
+                                  ClutterFrameInfo *frame_info,
+                                  gint64            presentation_time)
 {
   MetaWindowActorPrivate *priv = self->priv;
   GList *l;
@@ -2105,7 +2076,7 @@ meta_window_actor_frame_complete (MetaWindowActor *self,
     {
       GList *l_next = l->next;
       FrameData *frame = l->data;
-      gint64 frame_counter = cogl_frame_info_get_frame_counter (frame_info);
+      gint64 frame_counter = frame_info->frame_counter;
 
       if (frame->frame_counter != -1 && frame->frame_counter <= frame_counter)
         {
